@@ -42,6 +42,12 @@ def _format_http_method_count(data):
 
 def _escape(value):
     if value is not None:
+        if not isinstance(value, unicode):
+            # If a random byte string gets here, ensure it's ascii. Latin-1 is
+            # used to decode so the escaped form matches the original bytes.
+            value = value.decode("latin1").encode("ascii", "backslashreplace")
+        else:
+            value = value.encode("utf-8")
         value = cgi.escape(value)
     return value
 
@@ -656,6 +662,7 @@ class ErrorSummary:
         fp.write('<html>\n'
                  '<head>\n'
                  '<title>Oops Report Summary</title>\n'
+                 '<meta http-equiv="Content-Type" content="text/html;charset=UTF-8" />\n'
                  '<link rel="stylesheet" type="text/css" href="%s/oops/static/oops.css" />\n'
                  '</head>\n'
                  '<body>\n'
