@@ -105,7 +105,7 @@ class OopsReadError(IOError):
 class AppInstance(models.Model):
     """The app instance an oops belongs to."""
     # lpnet, edge, staging, etc.
-    title = models.CharField(unique=True, max_length=50)
+    title = models.CharField(unique=True, max_length=100)
 
     def __unicode__(self):
         return unicode(self.title)
@@ -113,13 +113,38 @@ class AppInstance(models.Model):
 
 class Prefix(models.Model):
     """Oops prefixes as defined by the app instance configuration."""
-    value = models.CharField(unique=True, max_length=20)
+    value = models.CharField(unique=True, max_length=100)
     appinstance = models.ForeignKey(AppInstance)
 
     def __unicode__(self):
         return unicode(self.value)
 
 
+class PruneInfo(models.Model):
+    """Information about oops pruning."""
+    
+    pruned_until = models.DateTimeField()
+
+    @classmethod
+    def prune_unreferenced(klass, prune_from, prune_until, references):
+        # XXX: This trusts date more than we may want to - should consider
+        # having a date_received separate to the date generated.
+        while True:
+            # There may be very many OOPS to prune, so only ask for a few at a
+            # time. This prevents running out of memory in the prune process
+            # (can happen when millions of reports are being pruned at once).
+            to_delete = set(Oops.objects.filter(
+                date__gte=prune_from, date__lte=prune_until).exclude(
+                    oopsid__in=references)[:10000])
+            # deleting 1 at a time is a lot of commits, but leaves the DB lively
+            # for other transactions. May need to batch this in future if its
+            # too slow.
+            for oopsid in to_delete:
+                Oops.objects.filter(oopsid__exact=oopsid).delete()
+            if not to_delete:
+                break
+
+
 class Report(models.Model):
     """A summary report for OOPSes of a set of prefixes."""
 
@@ -281,10 +306,18 @@ def _get_oops_tuple(oops):
     if not prefix:
         # Legacy support for pre-reporter using OOPSes.
         prefix_match = oops_re.match(oopsid)
-        if prefix_match is not None:
+        if prefix_match:
             prefix = prefix_match.group('oopsprefix')
-        else:
+            # Ubuntu One format is prefixZuuid.
+            if len(prefix) > 33 and prefix[-33] == 'Z':
+                prefix = prefix[:-33]
+            elif len(oopsid) >= 37:
+                # 37 is len (OOPS-<hash>).
+                prefix = None
+
+        if not prefix:
             prefix = 'UNKNOWN'
+
     prefix = prefix.upper()
     try:
         prefix = Prefix.objects.get(value__exact=prefix)
@@ -304,8 +337,34 @@ def _get_oops_tuple(oops):
         user_agent = 'No user agent',
         is_local_referrer = False)
     # Grab data needed by the Oops database model from the req_vars.
-    req_vars = oops.get('req_vars') or ()
-    for key, value in req_vars:
+    req_vars = oops.get('req_vars') or {}
+    # Some badly written OOPSes are tuples with single item tuples. (field.blob
+    # was seen).
+    # New ones are dicts.
+    def ensure_dict(iterable_or_dict):
+        try:
+            items = iterable_or_dict.items()
+            return iterable_or_dict
+        except AttributeError:
+            result = {}
+            for item in iterable_or_dict:
+                try:
+                    key, value = item
+                except ValueError:
+                    key = item[0]
+                    value = ''
+                result[key] = value
+            return result
+    req_vars = ensure_dict(req_vars)
+    for key, value in req_vars.items():
+        if isinstance(value, str):
+            try:
+                # We can get anything in HTTP headers
+                # Trying to squeeze the bytes into ASCII plane
+                value.decode('ascii')
+            except UnicodeDecodeError:
+                # If that fails, percent-quote them
+                value = urllib.quote(value)
         if key == 'REQUEST_METHOD':
             # Truncate the REQUEST_METHOD if it's longer than 10 characters
             # since a str longer than that is not allowed in the DB.
@@ -320,6 +379,14 @@ def _get_oops_tuple(oops):
         elif key == 'HTTP_USER_AGENT':
             data['user_agent'] = conform(value, 200)
     statements = oops.get('timeline') or []
+    # Sanity check/coerce the statements into what we expect.
+    # Must be a list:
+    if not isinstance(statements, list):
+        statements = [(0, 0, 'badtimeline', str(statements), 'unknown')]
+    else:
+        filler = (0, 0, 'unknown', 'unknown', 'unknown')
+        for row, action in enumerate(statements):
+            statements[row] = tuple(action[:5]) + filler[len(action):]
     duration = oops.get('duration')
     if duration is not None:
         total_time = int(duration * 1000)
@@ -353,7 +420,7 @@ def _get_oops_tuple(oops):
         most_expensive_statement = None
     else:
         costs = {}
-        for (start, end, db_id, statement) in statements:
+        for (start, end, db_id, statement, tb) in statements:
             statement = replace_variables(statement)
             cost = end - start
             if statement in costs:
@@ -403,7 +470,8 @@ def _get_oops_tuple(oops):
         is_bot = _robot_pat.search(data['user_agent']) is not None,
         statements_count = len(statements),
         )
-    return data, req_vars, statements, oops.get('tb_text') or ''
+    return (
+        data, sorted(req_vars.items()), statements, oops.get('tb_text') or '')
 
 
 def parsed_oops_to_model_oops(parsed_oops, pathname):
@@ -487,6 +555,11 @@ class Oops(models.Model):
         self.parsed_oops = _get_oops(self.pathname)
         return self.parsed_oops
 
+    @readproperty
+    def hostname(self):
+        # Currently from oops reports only.
+        return self.parsed_oops.get('hostname', 'Unknown') or 'Unknown'
+
     def _set_values(self):
         ignore, self.req_vars, self.statements, self.traceback = (
             _get_oops_tuple(self.parsed_oops))
@@ -530,8 +603,8 @@ class Oops(models.Model):
         #XXX need docstring and test
         if not self.statements:
             return []
-        return [(count+1, start, stop, db_id, stop - start, stmt) for
-                (count, (start, stop, db_id, stmt)) in
+        return [(count+1, start, stop, db_id, stop - start, stmt, tb) for
+                (count, (start, stop, db_id, stmt, tb)) in
                 enumerate(self.statements)]
 
     def repeated_statements(self):
@@ -587,7 +660,7 @@ class Oops(models.Model):
         if not self.statements:
             return 0
         accumulated_time = 0
-        for (start, stop, db_id, stmt) in self.statements:
+        for (start, stop, db_id, stmt, trace) in self.statements:
             length = stop - start
             accumulated_time += length
         return accumulated_time
@@ -600,13 +673,13 @@ class Oops(models.Model):
     def _group_queries(self, unsorted_statements):
         """Group SQL log queries.
 
-        unsorted_statements is a list of (start, stop, db_id, statement).
+        unsorted_statements is a list of (start, stop, db_id, statement, tb).
         Return a list of (total_time_ms, count, mean_length, saving, db_id,
         statement).
         """
         sorted_statements = sorted(
             unsorted_statements,
-            key=lambda (start, stop, db_id, statement): statement
+            key=lambda row: row[3]
             )
         groups = []
         def append_to_group(gtime, gcount, db_id, statement):
@@ -617,7 +690,7 @@ class Oops(models.Model):
 
         group_statement = None
         for row in sorted_statements:
-            start, stop, db_id, statement = row
+            start, stop, db_id, statement, tb = row
             statement = replace_variables(statement)
             length = stop - start
             if statement != group_statement:
